(*********************************************************)
(* TSLXFilter is a data filter component for Delphi6     *)
(* distributed for free, non-commercial purpose          *)
(*  2007 by Sl-Alex http://www.sl-alex.nm.ru/           *)
(*********************************************************)
unit Filter;

interface

uses
  Classes, Math, FilterData;

type

//  
TSLXWindowType = (wt_SIN3, wt_RECTANGLE, wt_TRIANGLE, wt_CUSTOM);

// ,  ,       
TSLXCustomCoeffEvent = procedure(var NewUserCoeff: Double; const AtIndex: Integer) of Object;

//  
TSLXFilterWindow = class(TPersistent)
  private
    FOwner: TComponent;
    FCoeffsList: TList;                                         //  
    FWindowType: TSLXWindowType;                                //  
    FWindowCoeff: Double;                                       //   
    FOnGetCustomCoeff: TSLXCustomCoeffEvent;                    // 
    function FGetCoeffs(const AIndex: Integer): Double;         //  
    procedure FSetCoeffs(const AIndex: Integer; Value: Double); //  
    procedure FSetWindowType(Value: TSLXWindowType);            //   
    function FGetCoeffsCount: Integer;                          //   
    procedure FAddCoeff(Value: Double);                         //  
    procedure FClear;                                           //   (  )
  public
    constructor Create(AOwner: TComponent);
    destructor Destroy; override;
    property Coeffs[const AIndex: Integer]: Double read FGetCoeffs write FSetCoeffs; //  
    procedure RefreshCoeffs;                        //   
    property WindowCoeff: Double read FWindowCoeff; //   
  published
    property WindowType: TSLXWindowType read FWindowType write FSetWindowType; //  
    property Count: Integer read FGetCoeffsCount;                              //  
    property OnGetCustomCoeff: TSLXCustomCoeffEvent read FOnGetCustomCoeff write FOnGetCustomCoeff; // 
end;

// 
TSLXFilterPoint = class(TPersistent)
  private
    FValue: Double;
  public
    constructor Create(NewValue: Double);
    destructor Destroy; override;
    property Value: Double read FValue write FValue;
  published
end;

// 
  TSLXFilter = class(TComponent)
  private
    { Private declarations }
    FLength: Integer;                   //  
    FOut: Real;                         //  
    FIndex: Integer;                    //   
    FEnabled: Boolean;                  //   
    FWindow: TSLXFilterWindow;          //  
    FData: TSLXFilterData;              //  
    procedure FSetWindow(NewWindow: TSLXFilterWindow);   //   
    procedure FSetData(NewData: TSLXFilterData);         //  
    procedure FSetLength(NewLength: Integer);            //   
    function FGetOnGetCustomCoeff: TSLXCustomCoeffEvent; //  
    procedure FSetOnGetCustomCoeff(NewEvent: TSLXCustomCoeffEvent); //  
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create( AOwner: TComponent ); override;      //  
    destructor Destroy; override;                            //  
    procedure AddSample(Sample: Double);                     //  
    procedure Clear;                                         //  
    property FilterOut: Real read FOut;                      //  
    property Data: TSLXFilterData read FData write FSetData; //  
    procedure CalcOut;                                       //  
  published
    { Published declarations }
    property Length: Integer read FLength write FSetLength default 1;    //  
    property Enabled: Boolean read FEnabled write FEnabled default True; //   
    property Window: TSLXFilterWindow read FWindow write FSetWindow;     //  
    // ,  ,       
    property OnGetCustomCoeff: TSLXCustomCoeffEvent read FGetOnGetCustomCoeff write FSetOnGetCustomCoeff;
  end;

//  
procedure Register;

implementation

{$R Filter.dcr}

//   
procedure TSLXFilter.FSetLength(NewLength: Integer);
begin
  //  
  FWindow.FClear;
  FData.Clear;
  FLength:=NewLength;
  //  
  FWindow.RefreshCoeffs;
  //  
  FData.InitData(FOut);
end;

//   
procedure TSLXFilter.FSetWindow(NewWindow: TSLXFilterWindow);
begin
  FWindow:=NewWindow;
  FWindow.RefreshCoeffs;
end;

//   
procedure TSLXFilter.FSetData(NewData: TSLXFilterData);
begin
  FData:=NewData;
  FData.InitData(FOut);
end;

//     
procedure TSLXFilter.AddSample(Sample: Double);
begin
  if FEnabled then begin //   
    FData.AddData(Sample);
    CalcOut;
  end;
end;

//   (  )
procedure TSLXFilter.Clear;
begin
  FData.InitData(0);
  FIndex:=1;
end;

//   
procedure TSLXFilter.CalcOut;
var i: Integer;
begin
  //   
  FOut:=0;
  //    
  for i:=0 to FLength-1 do begin
    FOut:=FOut+FData.Items[i]*FWindow.Coeffs[i];
  end;
  if (FLength<>0) and (FWindow.WindowCoeff<>0) then FOut:=FOut/(FLength*FWindow.WindowCoeff);
end;

//  
function TSLXFilter.FGetOnGetCustomCoeff: TSLXCustomCoeffEvent;
begin
  FGetOnGetCustomCoeff:=Window.OnGetCustomCoeff;
end;

//  
procedure TSLXFilter.FSetOnGetCustomCoeff(NewEvent: TSLXCustomCoeffEvent);
begin
  Window.OnGetCustomCoeff:=NewEvent;
end;

//  
constructor TSLXFilter.Create( AOwner: TComponent );
begin
  inherited Create(AOwner);
  FLength:=10;
  FWindow:=TSLXFilterWindow.Create(Self);
  FData:=TSLXFilterData.Create(Self);
  FEnabled:=True;
end;

//  
destructor TSLXFilter.Destroy;
begin
  FData.Destroy;
  FWindow.Destroy;
  inherited Destroy;
end;

//----------------------------------------------
//  
//----------------------------------------------

//  
function TSLXFilterWindow.FGetCoeffs(const AIndex: Integer): Double;
begin
  FGetCoeffs:=TSLXFilterPoint(FCoeffsList.Items[AIndex]).Value;
end;

//  
procedure TSLXFilterWindow.FSetCoeffs(const AIndex: Integer; Value: Double);
begin
  TSLXFilterPoint(FCoeffsList.Items[AIndex]).Value:=Value;
end;

//   
procedure TSLXFilterWindow.FSetWindowType(Value: TSLXWindowType);
begin
  FWindowType:=Value;
end;

//   
function TSLXFilterWindow.FGetCoeffsCount;
begin
  FGetCoeffsCount:=FCoeffsList.Count;
end;

//  
procedure TSLXFilterWindow.FAddCoeff(Value: Double);
begin
  FCoeffsList.Add(TSLXFilterPoint.Create(Value));
end;

//    
procedure TSLXFilterWindow.RefreshCoeffs;
var FilterLen,i: Integer;
          TempCoeff: Double;
begin
  FilterLen:=(FOwner as TSLXFilter).Length;
  FClear;
  for i:=0 to FilterLen-1 do begin
    case FWindowType of
     wt_SIN3:      begin
                     FAddCoeff(IntPower(Sin((PI*(i+1))/(FilterLen+1)),3));
                   end;
     wt_RECTANGLE: begin
                     FAddCoeff(1);
                   end;
     wt_TRIANGLE:  begin
                     FAddCoeff(1-abs(((FilterLen-1)/2-i)/(FilterLen/2)));
                   end;
     wt_CUSTOM:  begin
                     if Assigned(OnGetCustomCoeff) then begin
                       FOnGetCustomCoeff(TempCoeff,i);
                     end else begin
                       TempCoeff:=0;
                     end;
                     FAddCoeff(TempCoeff);
                   end;
    end;
  end;
  //    
  FWindowCoeff:=0;
  for i:=0 to FilterLen-1 do begin
    FWindowCoeff:=FWindowCoeff+Coeffs[i];
  end;
  if FilterLen<>0 then
    FWindowCoeff:=FWindowCoeff/FilterLen;
end;

//   (  )
procedure TSLXFilterWindow.FClear;
var i: Integer;
begin
  //   
  for i:=0 to FCoeffsList.Count-1 do begin
    TSLXFilterPoint(FCoeffsList.Items[i]).Destroy;
  end;
  FCoeffsList.Clear;
end;

constructor TSLXFilterWindow.Create(AOwner: TComponent);
begin
  inherited Create;
  FOwner:=AOwner;
  FCoeffsList:=TList.Create;
  FWindowType:=wt_SIN3;
  RefreshCoeffs;
end;

destructor TSLXFilterWindow.Destroy;
begin
  FCoeffsList.Destroy;
  inherited Destroy;
end;

constructor TSLXFilterPoint.Create(NewValue: Double);
begin
  inherited Create;
  FValue:=NewValue;
end;

destructor TSLXFilterPoint.Destroy;
begin
  inherited Destroy;
end;

//     
procedure Register;
begin
  RegisterComponents('Sl-Alex', [TSLXFilter]);
end;

end.
